from django.utils.translation import gettext as _
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)

from wbhuman_resources.serializers import IncreaseDaySerializer


class AbsenceRequestButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbhuman_resources:absencerequest",),
                key="increase_days",
                action_label=_("Increasing Days"),
                label=_("Increase Days"),
                icon=WBIcon.ADD.icon,
                title=_("Increase Days"),
                serializer=IncreaseDaySerializer,
                description_fields=_(
                    """
                    <p>You are about to increase the abscence request by {{number_days}} days</p>
                """
                ),
                instance_display=create_simple_display([["number_days"]]),
            )
        }



import zoneinfo
from datetime import date, timedelta
from unittest.mock import patch

import pandas as pd
import pytest
from django.contrib.auth.models import Group, Permission
from django.db.models import Sum
from faker import Faker
from psycopg.types.range import TimestamptzRange

from wbhuman_resources.models import (
    AbsenceRequest,
    AbsenceRequestPeriods,
    AbsenceRequestType,
    BalanceHourlyAllowance,
)
from wbhuman_resources.models.absence import (
    can_cancel_request,
    can_validate_or_deny_request,
)
from wbhuman_resources.models.preferences import (
    get_previous_year_balance_expiration_date,
)

fake = Faker()


@pytest.mark.django_db
class TestAbsenceRequest:
    @pytest.fixture()
    def past_aware_datetime(self):
        return fake.past_datetime().astimezone(zoneinfo.ZoneInfo("UTC"))

    @pytest.fixture()
    def future_aware_datetime(self):
        return fake.future_datetime().astimezone(zoneinfo.ZoneInfo("UTC"))

    def test_normal_user_cannot_cancel_past_request(self, vacation_request_factory, past_aware_datetime):
        request = vacation_request_factory(
            status=AbsenceRequest.Status.APPROVED,
            period=TimestamptzRange(lower=past_aware_datetime - timedelta(days=2), upper=past_aware_datetime),
        )
        assert not can_cancel_request(request, request.employee.profile.user_account)

    def test_normal_user_can_cancel_future_request(
        self, vacation_request_factory, user_factory, future_aware_datetime, employee_human_resource_factory
    ):
        admin = user_factory.create(is_active=True)
        admin.user_permissions.add(Permission.objects.get(codename="administrate_absencerequest"))
        manager = employee_human_resource_factory.create()

        employee = employee_human_resource_factory.create(direct_manager=manager.profile)
        request = vacation_request_factory(
            employee=employee,
            status=AbsenceRequest.Status.APPROVED,
            period=TimestamptzRange(lower=future_aware_datetime, upper=future_aware_datetime + timedelta(days=2)),
        )

        assert can_cancel_request(request, request.employee.profile.user_account)
        assert can_cancel_request(request, admin)
        assert can_cancel_request(request, manager.profile.user_account)

    def test_can_validate_or_deny_request(
        self, vacation_request_factory, user_factory, future_aware_datetime, employee_human_resource_factory
    ):
        admin = user_factory.create(is_active=True)
        admin.user_permissions.add(Permission.objects.get(codename="administrate_absencerequest"))
        manager = employee_human_resource_factory.create()

        employee = employee_human_resource_factory.create(direct_manager=manager.profile)
        request = vacation_request_factory(
            employee=employee,
            status=AbsenceRequest.Status.APPROVED,
            period=TimestamptzRange(lower=future_aware_datetime, upper=future_aware_datetime + timedelta(days=2)),
        )

        assert not can_validate_or_deny_request(request, request.employee.profile.user_account)
        assert can_validate_or_deny_request(request, admin)
        assert can_validate_or_deny_request(request, manager.profile.user_account)

    def test_submit(self, absence_request_factory, employee_human_resource):
        request = absence_request_factory.create(status=AbsenceRequest.Status.DRAFT)
        request.employee.direct_manager = employee_human_resource.profile
        request.employee.save()

        request.submit()
        request.save()
        assert request.status == AbsenceRequest.Status.PENDING

    def test_approve(self, absence_request_factory):
        request = absence_request_factory.create(status=AbsenceRequest.Status.PENDING)

        request.approve()
        request.save()
        assert request.status == AbsenceRequest.Status.APPROVED

    def test_deny_not_vacation_request(self, time_off_request_factory):
        request = time_off_request_factory.create(status=AbsenceRequest.Status.PENDING)
        request.deny()
        request.save()
        assert request.status == AbsenceRequest.Status.DENIED

    def test_deny_vacation_request(self, vacation_request_factory):
        request = vacation_request_factory.create(status=AbsenceRequest.Status.PENDING)
        request.deny()
        request.save()
        assert request.status == AbsenceRequest.Status.DENIED

    def test_backtodraft(self, absence_request_factory):
        request = absence_request_factory.create(status=AbsenceRequest.Status.PENDING)
        request.backtodraft()
        request.save()
        assert request.status == AbsenceRequest.Status.DRAFT

    def test_cancel(self, absence_request_factory, employee_human_resource):
        request = absence_request_factory.create(status=AbsenceRequest.Status.APPROVED)
        request.employee.direct_manager = employee_human_resource.profile
        request.employee.save()

        request.cancel()
        request.save()
        assert request.status == AbsenceRequest.Status.CANCELLED

    def test_timespan(self, absence_request_periods):
        request = absence_request_periods.request
        assert (
            request.periods_timespan.lower
            == request.periods.earliest("timespan__startswith").request.periods_timespan.lower
        )
        assert (
            request.periods_timespan.upper
            == request.periods.latest("timespan__startswith").request.periods_timespan.upper
        )

    def test_can_delete_draft_request(self, absence_request_factory, user):
        request = absence_request_factory(status=AbsenceRequest.Status.DRAFT)
        assert request.is_deletable_for_user(user)

    def test_can_delete_future_pending_request(self, absence_request_factory, future_aware_datetime, user):
        request = absence_request_factory(
            status=AbsenceRequest.Status.PENDING,
            period=TimestamptzRange(lower=future_aware_datetime, upper=future_aware_datetime + timedelta(days=2)),
        )
        assert request.is_deletable_for_user(user)

    def test_can_delete_past_pending_request(self, absence_request_factory, past_aware_datetime, user_factory):
        normal_user = user_factory.create()
        admin_user = user_factory.create()
        admin_user.user_permissions.add(Permission.objects.get(codename="administrate_absencerequest"))
        request = absence_request_factory.create(
            status=AbsenceRequest.Status.PENDING,
            period=TimestamptzRange(lower=past_aware_datetime - timedelta(days=2), upper=past_aware_datetime),
        )
        assert not request.is_deletable_for_user(normal_user)
        assert request.is_deletable_for_user(admin_user)

    # Property checks and test
    def test_total_hours(self, day_off_calendar, absence_request_periods_factory):
        morning = day_off_calendar.default_periods.earliest("lower_time")
        afternoon = day_off_calendar.default_periods.latest("lower_time")
        p1 = absence_request_periods_factory.create(default_period=morning)
        request = p1.request
        absence_request_periods_factory.create(request=request, default_period=afternoon, date=p1.date)
        assert request.total_hours == morning.total_hours + afternoon.total_hours
        assert AbsenceRequest.objects.get(id=request.id)._total_hours == morning.total_hours + afternoon.total_hours
        absence_request_periods_factory.create(
            request=request, default_period=morning, date=(p1.date + pd.tseries.offsets.BDay(1)).date()
        )
        absence_request_periods_factory.create(
            request=request, default_period=afternoon, date=(p1.date + pd.tseries.offsets.BDay(1)).date()
        )
        assert request.total_hours == (morning.total_hours + afternoon.total_hours) * 2
        assert (
            AbsenceRequest.objects.get(id=request.id)._total_hours == (morning.total_hours + afternoon.total_hours) * 2
        )

    def test_total_vacation_hours(self, absence_request_factory, absence_request_type_factory):
        time_off_request = absence_request_factory.create(
            status=AbsenceRequest.Status.APPROVED, type=absence_request_type_factory.create(is_vacation=False)
        )
        vacation_request = absence_request_factory.create(
            status=AbsenceRequest.Status.APPROVED, type=absence_request_type_factory.create(is_vacation=True)
        )
        unapprove_vacation_request = absence_request_factory.create(
            type=absence_request_type_factory.create(is_vacation=True)
        )

        assert time_off_request.total_vacation_hours == 0
        assert AbsenceRequest.objects.get(id=time_off_request.id)._total_vacation_hours == 0

        assert unapprove_vacation_request.total_vacation_hours == 0
        assert AbsenceRequest.objects.get(id=unapprove_vacation_request.id)._total_vacation_hours == 0

        assert vacation_request.total_vacation_hours
        assert vacation_request.total_vacation_hours == vacation_request.periods.aggregate(s=Sum("_total_hours"))["s"]
        assert (
            AbsenceRequest.objects.get(id=vacation_request.id)._total_vacation_hours
            == vacation_request.total_vacation_hours
        )

    def test_total_hours_in_days(self, absence_request):
        exp_res = (
            absence_request.periods.aggregate(s=Sum("_total_hours"))["s"]
            / absence_request.employee.calendar.get_daily_hours()
        )
        assert exp_res
        assert absence_request.total_hours_in_days == exp_res
        assert AbsenceRequest.objects.get(id=absence_request.id)._total_hours_in_days == exp_res

    def test_total_vacation_hours_in_days(self, absence_request_factory, absence_request_type_factory):
        vacation_request = absence_request_factory.create(
            status=AbsenceRequest.Status.APPROVED, type=absence_request_type_factory.create(is_vacation=True)
        )
        assert vacation_request.total_vacation_hours_in_days
        assert (
            vacation_request.total_vacation_hours_in_days
            == vacation_request.periods.aggregate(s=Sum("_total_hours"))["s"]
            / vacation_request.employee.calendar.get_daily_hours()
        )
        assert (
            AbsenceRequest.objects.get(id=vacation_request.id)._total_vacation_hours_in_days
            == vacation_request.total_vacation_hours_in_days
        )

    @patch("wbhuman_resources.models.absence.send_notification")
    def test_notify_requester(self, mock_fct, absence_request):
        title = fake.sentence()
        message = fake.sentence()
        absence_request.notify(title, message, to_requester=True)
        mock_fct.assert_called_with(
            code="wbhuman_resources.absencerequest.notify",
            title=title,
            body=message,
            reverse_name="wbhuman_resources:absencerequest-detail",
            reverse_args=[absence_request.id],
            user=absence_request.employee.profile.user_account,
        )

    @patch("wbhuman_resources.models.absence.send_notification")
    def test_notify_managers(self, mock_fct, absence_request, authenticated_person_factory):
        direct_manager = authenticated_person_factory.create()
        absence_request.employee.direct_manager = direct_manager
        absence_request.employee.save()

        # add a general manager
        general_manager = authenticated_person_factory.create()
        general_manager.user_account.user_permissions.add(
            Permission.objects.get(codename="administrate_employeehumanresource")
        )

        absence_request.refresh_from_db()

        title = fake.sentence()
        message = fake.sentence()
        absence_request.notify(title, message, to_requester=False, to_manager=True)
        mock_fct.assert_any_call(
            code="wbhuman_resources.absencerequest.notify",
            title=title,
            body=message,
            reverse_name="wbhuman_resources:absencerequest-detail",
            reverse_args=[absence_request.id],
            user=general_manager.user_account,
        )
        mock_fct.assert_any_call(
            code="wbhuman_resources.absencerequest.notify",
            title=title,
            body=message,
            reverse_name="wbhuman_resources:absencerequest-detail",
            reverse_args=[absence_request.id],
            user=direct_manager.user_account,
        )

    @patch("wbhuman_resources.models.absence.send_notification")
    def test_notify_extra_notify_user(self, mock_fct, absence_request, user):
        # We create a user, add it to a test group and add this group to the absence request type "extra_notify_group"
        group = Group.objects.create(name="test")
        user.groups.add(group)
        absence_request.type.extra_notify_groups.add(group)
        absence_request.refresh_from_db()

        title = fake.sentence()
        message = fake.sentence()
        absence_request.notify(title, message, to_requester=False, to_manager=True)
        mock_fct.assert_called_with(
            code="wbhuman_resources.absencerequest.notify",
            title=title,
            body=message,
            reverse_name="wbhuman_resources:absencerequest-detail",
            reverse_args=[absence_request.id],
            user=user,
        )


@pytest.mark.django_db
class TestAbsenceRequestType:
    def test_get_choices(self, absence_request_type_factory):
        t1 = absence_request_type_factory.create()
        t2 = absence_request_type_factory.create()
        assert AbsenceRequestType.get_choices() == [(t1.id, t1.title), (t2.id, t2.title)]

    def test_validate_country_needed_but_not_specified(self, absence_request_type_factory):
        absence_request_type = absence_request_type_factory.create(is_country_necessary=True)
        with pytest.raises(ValueError):
            absence_request_type.validate_country(None)

    def test_validate_country_specified_and_allowed(self, absence_request_type_factory, country):
        absence_request_type = absence_request_type_factory.create(
            is_country_necessary=True,
        )  # test that the post_save automatically appends the newly created country to the list of allowed countries
        assert absence_request_type.validate_country(country)

    def test_validate_country_specified_but_not_allowed(self, absence_request_type_factory, country):
        absence_request_type = absence_request_type_factory.create(is_country_necessary=True)
        absence_request_type.crossborder_countries.clear()
        with pytest.raises(ValueError):
            absence_request_type.validate_country(country)


@pytest.mark.django_db
class TestAbsenceRequestPeriod:
    def test_total_hours(self, absence_request_periods):
        exp_res = absence_request_periods.default_period.total_hours
        assert exp_res
        assert absence_request_periods.total_hours == exp_res
        assert AbsenceRequestPeriods.objects.get(id=absence_request_periods.id)._total_hours == exp_res

    @pytest.mark.parametrize("past_date,future_date", [(fake.past_date(), fake.future_date())])
    def test_previous_vacation_period(self, absence_request_periods_factory, past_date, future_date):
        current = absence_request_periods_factory.create(date=date.today())
        past = absence_request_periods_factory.create(employee=current.employee, date=past_date)
        future = absence_request_periods_factory.create(employee=current.employee, date=future_date)
        AbsenceRequest.objects.update(status=AbsenceRequest.Status.APPROVED, type=current.request.type)
        assert current.previous_period == past
        assert future.previous_period == current
        assert past.previous_period is None

    @pytest.mark.parametrize("test_date", [(fake.date_this_year())])
    def test_get_periods_as_df(
        self, day_off_calendar, employee_human_resource_factory, absence_request_periods_factory, test_date
    ):
        employee1 = employee_human_resource_factory.create(calendar=day_off_calendar, is_active=True)
        employee2 = employee_human_resource_factory.create(
            calendar=day_off_calendar, is_active=False
        )  # Expect this employee to not be present
        p1 = absence_request_periods_factory.create(employee=employee1, date=test_date)
        p2 = absence_request_periods_factory.create(employee=employee2, date=test_date)
        p3 = absence_request_periods_factory.create(employee=employee1, date=test_date - timedelta(days=1))
        res = AbsenceRequestPeriods.get_periods_as_df(
            test_date, test_date + timedelta(days=1), employee__is_active=True
        )
        assert res.shape == (1, 5)
        res = res.set_index(["employee", "period", "date"])
        assert res.loc[(employee1.id, p1.default_period.id, test_date), :].values.tolist() == [
            p1.request.type.title,
            p1.request.status,
        ]

        with pytest.raises(KeyError):
            assert res.loc[(employee2.id, p2.default_period.id, test_date), :].values.tolist() == [
                p2.request.type.title,
                p2.request.status,
            ]
            assert res.loc[(employee1.id, p3.period.id, test_date - timedelta(days=1)), :]

    @pytest.mark.parametrize("year_str", [(fake.year())])
    def test_assign_balance(
        self, employee_human_resource, absence_request_periods_factory, employee_year_balance_factory, year_str
    ):
        year = int(year_str)
        previous_balance = employee_year_balance_factory.create(
            employee=employee_human_resource, extra_balance=0, year=year - 1
        )
        current_balance = employee_year_balance_factory.create(
            employee=employee_human_resource, extra_balance=0, year=year
        )
        next_balance = employee_year_balance_factory.create(
            employee=employee_human_resource, extra_balance=0, year=year + 1
        )
        BalanceHourlyAllowance.objects.update(
            hourly_allowance=4
        )  # Update all created balance allowance with 4 crdits (corresponds to a period)

        p1 = absence_request_periods_factory.create(
            balance=None, employee=employee_human_resource, date=fake.date_between(date(year, 1, 1), date(year, 3, 31))
        )
        p1.assign_balance()
        assert p1.balance == previous_balance

        p2 = absence_request_periods_factory.create(
            balance=None,
            employee=employee_human_resource,
            date=fake.date_between(date(year, 3, 31), date(year, 6, 30)),
        )
        p2.assign_balance()
        assert p2.balance == current_balance

        p3 = absence_request_periods_factory.create(
            balance=None,
            employee=employee_human_resource,
            date=fake.date_between(date(year, 6, 30), date(year, 12, 31)),
        )
        p3.assign_balance()
        assert p3.balance == next_balance

        p4 = absence_request_periods_factory.create(  # request but no balance left
            balance=None,
            employee=employee_human_resource,
            date=fake.date_between(date(year, 1, 1), date(year, 12, 31)),
        )
        p4.assign_balance()
        assert p4.balance.year == year + 2

    @pytest.mark.parametrize("year_str", [(fake.year())])
    def test_assign_balance_with_expired_balance(
        self, employee_human_resource, absence_request_periods_factory, employee_year_balance_factory, year_str
    ):
        year = int(year_str)
        expiration_date = get_previous_year_balance_expiration_date(year)

        for y in range(year, expiration_date.year - 1):
            employee_year_balance_factory.create(employee=employee_human_resource, extra_balance=0, year=y)
        BalanceHourlyAllowance.objects.filter(balance__year=year).update(
            hourly_allowance=4
        )  # Update all created balance allowance with 4 crdits (corresponds to a period)
        BalanceHourlyAllowance.objects.exclude(balance__year=year).update(
            hourly_allowance=0
        )  # and set anything else to 0
        p1 = absence_request_periods_factory.create(
            balance=None,
            employee=employee_human_resource,
            date=fake.date_between(expiration_date, date(expiration_date.year, 12, 31)),
        )
        p1.assign_balance()
        assert p1.balance is not None
        assert p1.balance.year == year + 1
        assert p1.balance.balance == 0

    def test_no_vacation_or_approved_request_has_no_balance(self, absence_request_periods):
        AbsenceRequest.objects.update(status=AbsenceRequest.Status.DRAFT)
        AbsenceRequestType.objects.update(is_vacation=False)
        AbsenceRequestPeriods.objects.update(balance=None)
        absence_request_periods.refresh_from_db()
        absence_request_periods.assign_balance()
        assert absence_request_periods.balance is None

    def test_get_consecutive_hours_count(
        self,
        absence_request_periods_factory,
        employee_human_resource,
        employee_weekly_off_periods_factory,
        day_off_factory,
    ):
        morning = employee_human_resource.calendar.default_periods.earliest("lower_time")
        afternoon = employee_human_resource.calendar.default_periods.latest("lower_time")
        hours_per_period = 4
        p1 = absence_request_periods_factory.create(employee=employee_human_resource, default_period=morning)
        assert p1.consecutive_hours_count == hours_per_period
        p2 = absence_request_periods_factory.create(
            date=p1.date, employee=employee_human_resource, default_period=afternoon
        )  # Straight next period, counter should be incremented
        assert p2.consecutive_hours_count == hours_per_period * 2
        employee_weekly_off_periods_factory.create(
            period=morning, weekday=(p1.date.weekday() + 1) % 6, employee=employee_human_resource
        )
        p3 = absence_request_periods_factory.create(
            date=p1.date + timedelta(days=1), employee=employee_human_resource, default_period=afternoon
        )  # Expected to keep incremeting counter because the previous employee's weekly day off period is jumped
        assert p3.consecutive_hours_count == hours_per_period * 3
        day_off_factory.create(calendar=employee_human_resource.calendar, date=p1.date + timedelta(days=2))
        p4 = absence_request_periods_factory.create(
            date=p1.date + timedelta(days=3), employee=employee_human_resource, default_period=morning
        )  # Expected to keep incremeting counter because the previous day off is jumped
        assert p4.consecutive_hours_count == hours_per_period * 4
        p5 = absence_request_periods_factory.create(
            date=p1.date + timedelta(days=4), employee=employee_human_resource, default_period=morning
        )
        assert p5.consecutive_hours_count == hours_per_period  # Expect  reset of counter



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbhuman_resources.models.kpi import KPI


class KPIDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="handler", label=_("KPI")),
                dp.Field(key="evaluated_persons", label=_("Persons")),
                dp.Field(key="period", label=_("Period")),
                dp.Field(key="parameters", label=_("Parameters")),
                dp.Field(key="is_active", label=_("Is Active")),
            ]
        )

    def get_instance_display(self) -> Display:
        sections = []
        grid_fields = [
            [repeat_field(2, "name")],
            [repeat_field(2, "handler")],
            [repeat_field(2, "evaluated_persons")],
            ["evaluated_intervals", "is_active"],
            ["goal", "individual_evaluation"],
            ["period", "last_update"],
        ]
        if pk := self.view.kwargs.get("pk", None):
            handler = KPI.objects.get(id=pk).get_handler()
            sections = [
                create_simple_section("parameters_section", _("Parameters"), handler.get_display_grid()),
                create_simple_section("evaluation_section", _("Evaluations"), [["evaluations"]], "evaluations"),
            ]
            grid_fields.extend([["parameters_section", "evaluation_section"]])
        return create_simple_display(grid_fields, sections)


class KPIEvaluationDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="person", label=_("Person")),
                dp.Field(key="evaluated_score", label=_("Evaluated Score")),
                dp.Field(key="goal", label=_("Goal")),
                dp.Field(key="evaluation_date", label=_("Evaluation Date")),
                dp.Field(key="evaluated_period", label=_("Evaluated Period")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "kpi")],
                [repeat_field(2, "person")],
                [repeat_field(2, "evaluated_period")],
                ["evaluation_date", "last_update"],
                [repeat_field(2, "evaluated_score")],
            ]
        )


class KPIEvaluationPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        pandas_fields = [
            dp.Field(key="person", label=_("Person")),
            dp.Field(key="kpi_name", label=_("KPI")),
            dp.Field(key="period", label=_("Period")),
            dp.Field(key="evaluation_date", label=_("Evaluation Date")),
            dp.Field(key="evaluated_score", label=_("Evaluated Score")),
            dp.Field(
                key="goal",
                label=_("Goal"),
                formatting_rules=[
                    dp.FormattingRule(
                        style={"fontWeight": "bold"},
                    ),
                ],
            ),
            dp.Field(key="progression", label=_("Progression")),
        ]
        return dp.ListDisplay(fields=pandas_fields)



from typing import TYPE_CHECKING, Optional

from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Operator, Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Style,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbhuman_resources.models import Review, ReviewQuestion

if TYPE_CHECKING:
    from wbhuman_resources.viewsets.review import ReviewAnswerReviewPandasViewSet


def get_legends(model=None):
    list_format = []
    if model:
        for _status, color in model.Status.get_color_map():
            list_format.append(dp.LegendItem(icon=color, label=_status.label, value=_status.value))
    return [dp.Legend(key="status", items=list_format)]


def get_list_formatting(model=None):
    color_conditions = []
    if model:
        for _status, color in model.Status.get_color_map():
            color_conditions.append(
                dp.FormattingRule(condition=("==", _status.name), style={"backgroundColor": color})
            )
    return [
        dp.Formatting(column="status", formatting_rules=color_conditions),
    ]


class ReviewGroupDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["name", "employees"], [repeat_field(2, "review_section")]],
            [create_simple_section("review_section", _("Review"), [["review"]], "review", collapsed=False)],
        )


class ReviewDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="year", label=gettext_lazy("Year"), width=Unit.PIXEL(80)),
                dp.Field(key="type", label=gettext_lazy("Type"), width=Unit.PIXEL(100)),
                dp.Field(key="reviewee", label=gettext_lazy("Reviewee"), width=Unit.PIXEL(150)),
                dp.Field(key="reviewer", label=gettext_lazy("Reviewer"), width=Unit.PIXEL(150)),
                dp.Field(key="moderator", label=gettext_lazy("Moderator"), width=Unit.PIXEL(150)),
                dp.Field(key="from_date", label=gettext_lazy("From"), width=Unit.PIXEL(140)),
                dp.Field(key="to_date", label=gettext_lazy("To"), width=Unit.PIXEL(140)),
                dp.Field(key="review_deadline", label=gettext_lazy("Deadline"), width=Unit.PIXEL(140)),
                dp.Field(key="status", label=gettext_lazy("Status"), width=Unit.PIXEL(150)),
                dp.Field(key="review", label=gettext_lazy("Review date"), width=Unit.PIXEL(140)),
                dp.Field(key="is_template", label=gettext_lazy("Is Template"), width=Unit.PIXEL(100)),
                dp.Field(key="changed", label=gettext_lazy("Changed"), width=Unit.PIXEL(140)),
                dp.Field(key="review_group", label=gettext_lazy("Group"), width=Unit.PIXEL(150)),
            ],
            legends=get_legends(Review),
            formatting=get_list_formatting(Review),
        )

    def get_main_information_page(self):
        grid_areas = [
            ["year", "type", repeat_field(2, "status")],
            ["review_group", "is_template", "moderator", "auto_apply_deadline"],
        ]
        if self.view.kwargs.get("pk"):
            instance = self.view.get_object()
            if not instance.is_template:
                grid_areas = [
                    ["year", "type", repeat_field(2, "status")],
                    ["from_date", "to_date", "review_deadline", "review"],
                    ["reviewee", "reviewer", "moderator", "auto_apply_deadline"],
                ]
            if instance.status != Review.Status.PREPARATION_OF_REVIEW:
                if self.request.user.profile in [instance.reviewer, instance.reviewee]:
                    grid_areas.append(
                        ["completely_filled_reviewee", "completely_filled_reviewer", repeat_field(2, "global_rating")]
                    )
                else:
                    grid_areas.append(
                        ["completely_filled_reviewee", "completely_filled_reviewer", repeat_field(2, ".")]
                    )

                if instance.status in [Review.Status.EVALUATION, Review.Status.VALIDATION]:
                    grid_areas.append([repeat_field(2, "signed_reviewee"), repeat_field(2, "signed_reviewer")])
                    grid_areas.append([repeat_field(2, "feedback_reviewee"), repeat_field(2, "feedback_reviewer")])

        return Page(
            title=_("Main Information"),
            layouts={
                default(): Layout(
                    grid_template_areas=grid_areas,
                ),
            },
        )

    def get_answers_page(self, instance):
        nb_columns = 4
        grid_areas = []
        sections = []
        inlines = []
        grid_template_rows = []
        if instance.status == Review.Status.PREPARATION_OF_REVIEW:
            grid_areas.append([repeat_field(nb_columns, "category_section")])
            sections.append(
                create_simple_section(
                    "category_section", _("Question Categories"), [["category"]], "category", collapsed=True
                )
            )
            categories = instance.get_question_categories()
            total, count = categories.count() + 1, 0
            for count, category in enumerate(categories, start=1):
                category_key = f"questioncategory{category.id}"
                grid_areas.append([repeat_field(nb_columns, f"{category_key}_section")])
                sections.append(
                    create_simple_section(
                        f"{category_key}_section",
                        ("({count}/{total}). {title}").format(count=count, total=total, title=category.name),
                        [[category_key]],
                        category_key,
                        collapsed=True,
                    )
                )

            grid_areas.append([repeat_field(nb_columns, "questionnocategory_section")])
            sections.append(
                create_simple_section(
                    "questionnocategory_section",
                    _("({count}/{total}). No Question Category").format(count=count + 1, total=total),
                    [["questionnocategory"]],
                    "questionnocategory",
                    collapsed=False,
                )
            )

        elif instance.status in [Review.Status.FILL_IN_REVIEW, Review.Status.REVIEW]:
            categories = instance.get_answer_categories_for_user(self.request.user)
            total, count = categories.count(), 0
            if no_category := ReviewQuestion.objects.filter(review=instance, category=None).exists():
                total += 1

            for count, category in enumerate(categories, start=1):
                category_key = f"reviewanswerquestioncategory{category.id}"
                grid_areas.append([repeat_field(nb_columns, f"{category_key}_section")])
                sections.append(
                    create_simple_section(
                        f"{category_key}_section",
                        ("({count}/{total}). {title}").format(count=count, total=total, title=category.name),
                        [[category_key]],
                        category_key,
                        collapsed=True,
                    )
                )

            if no_category:
                grid_areas.append([repeat_field(nb_columns, "reviewanswerquestionnocategory_section")])
                sections.append(
                    create_simple_section(
                        "reviewanswerquestionnocategory_section",
                        _("({count}/{total}). No Question Category").format(count=count + 1, total=total),
                        [["reviewanswerquestionnocategory"]],
                        "reviewanswerquestionnocategory",
                        collapsed=False,
                    )
                )

        elif instance.status in [Review.Status.EVALUATION, Review.Status.VALIDATION]:
            grid_areas += [
                [repeat_field(nb_columns, "rating_review_answer_section")],
                [repeat_field(nb_columns, "text_review_answer_inline")],
            ]
            sections.append(
                create_simple_section(
                    key="rating_review_answer_section",
                    title=_("Review of questions to be answered on a scale of 1 to 4"),
                    grid_template_areas=[["rating_review_answer_key"]],
                    inline_key="rating_review_answer_key",
                    collapsed=False,
                )
            )
            inlines.append(Inline(key="text_review_answer_inline", endpoint="text_review_answer_key"))
            grid_template_rows += [Style.MAX_CONTENT, Style.fr(1)]

        return Page(
            title=_("Q&A"),
            layouts={
                default(): Layout(
                    grid_template_areas=grid_areas,
                    sections=sections,
                    inlines=inlines,
                    grid_template_rows=grid_template_rows,
                ),
            },
        )

    def get_instance_display(self) -> Display:
        pages = []
        if self.view.kwargs.get("pk"):
            instance = self.view.get_object()
            pages.append(self.get_answers_page(instance))
        pages.append(self.get_main_information_page())
        return Display(pages=pages)


class ReviewTemplateDisplayConfig(ReviewDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="year", label=_("Year")),
                dp.Field(key="type", label=_("Type")),
                dp.Field(key="moderator", label=_("Moderator")),
                dp.Field(key="auto_apply_deadline", label=_("Auto Apply Deadline")),
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="review", label=_("Review")),
                dp.Field(key="review_group", label=_("Group")),
                dp.Field(key="changed", label=_("Changed")),
            ],
            formatting=get_list_formatting(Review),
        )


class ReviewQuestionCategoryDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name"), width=Unit.PIXEL(1000)),
                dp.Field(key="weight", label=_("Weight"), width=Unit.PIXEL(150)),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["name", "weight"], [repeat_field(2, "question_section")]],
            [
                create_simple_section(
                    "question_section", _("Question"), [["reviewquestion"]], "reviewquestion", collapsed=False
                )
            ],
        )


class ReviewQuestionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="question", label=gettext_lazy("Question"), width=Unit.PIXEL(600)),
                dp.Field(key="answer_type", label=gettext_lazy("Type"), width=Unit.PIXEL(200)),
                dp.Field(key="mandatory", label=gettext_lazy("Mandatory"), width=Unit.PIXEL(200)),
                dp.Field(
                    key="review_for",
                    label=gettext_lazy("Reviewee/ Reviewer/ Department/ Company"),
                    width=Unit.PIXEL(260),
                ),
                dp.Field(key="weight", label=gettext_lazy("Weight")),
                dp.Field(key="category", label=gettext_lazy("Category"), width=Unit.PIXEL(200), hide=True),
                dp.Field(key="review", label=gettext_lazy("Review"), hide=True),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["answer_type", "category", "review"],
                ["mandatory", "for_reviewee", "for_reviewer"],
                ["weight", "for_department_peers", "for_company_peers"],
                [repeat_field(3, "question")],
            ]
        )


class ReviewQuestionReviewDisplayConfig(ReviewQuestionDisplayConfig):
    def get_instance_display(self) -> Display:
        if self.view.review.status in [Review.Status.EVALUATION, Review.Status.VALIDATION]:
            return create_simple_display(
                [["review_answers_table"]],
                inlines=[Inline(key="review_answers_table", endpoint="review_answers_table")],
            )
        return super().get_instance_display()


class ReviewAnswerDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="question_name", label=_("Question"), width=Unit.PIXEL(720)),
            dp.Field(key="mandatory", label=_("Mandatory"), width=Unit.PIXEL(80)),
        ]
        if self.view.get_queryset().filter(question__answer_type=ReviewQuestion.ANSWERTYPE.RATING):
            fields += [dp.Field(key="answer_number", label=_("Rating"), width=Unit.PIXEL(170))]

        fields += [dp.Field(key="answer_text", label=_("Comment"), width=Unit.PIXEL(720))]

        if review_id := self.view.kwargs.get("review_id"):
            review = Review.objects.get(id=review_id)
            if review.status in [Review.Status.EVALUATION, Review.Status.VALIDATION]:
                fields = [
                    dp.Field(key="question_name", label=_("Question"), width=Unit.PIXEL(740)),
                    dp.Field(key="answered_by", label=_("Answered By"), width=Unit.PIXEL(140)),
                    dp.Field(key="answer_text", label=_("Comment"), width=Unit.PIXEL(740)),
                ]
        return dp.ListDisplay(fields=fields)

    def get_instance_display(self) -> Display:
        grid_fields = [[repeat_field(2, "question_name")], ["mandatory", "."]]
        if self.view.kwargs.get("pk", None):
            instance = self.view.get_object()
            if instance.question.answer_type == ReviewQuestion.ANSWERTYPE.RATING:
                grid_fields += [[repeat_field(2, "answer_number")]]
        grid_fields += [[repeat_field(2, "answer_text")]]
        return create_simple_display(grid_fields)


class ReviewProgressReviewPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        _fields = [
            dp.Field(key="answered_by_name", label=_("Answered By")),
            dp.Field(key="progress", label=_("Progress")),
        ]
        return dp.ListDisplay(fields=_fields)


class ReviewAnswerReviewPandasDisplayConfig(DisplayViewConfig):
    view: "ReviewAnswerReviewPandasViewSet"

    def _get_formatting_rules(self):
        return [
            dp.Formatting(
                column="deviation",
                formatting_rules=[
                    dp.FormattingRule(
                        style={
                            "backgroundColor": WBColor.YELLOW_LIGHT.value,
                        },
                        condition=dp.Condition(operator=Operator.EQUAL, value="LESS"),
                    ),
                    dp.FormattingRule(
                        style={
                            "backgroundColor": WBColor.GREEN_LIGHT.value,
                        },
                        condition=dp.Condition(operator=Operator.EQUAL, value="EQUAL"),
                    ),
                    dp.FormattingRule(
                        style={
                            "backgroundColor": WBColor.RED_LIGHT.value,
                        },
                        condition=dp.Condition(operator=Operator.EQUAL, value="GREAT"),
                    ),
                ],
            )
        ]

    def _get_legends(self):
        return [
            dp.Legend(
                key="deviation",
                items=[
                    dp.LegendItem(
                        icon=WBColor.RED_LIGHT.value,
                        label=_("Great Difference"),
                        value="GREAT",
                    ),
                    dp.LegendItem(
                        icon=WBColor.YELLOW_LIGHT.value,
                        label=_("Little Difference"),
                        value="LESS",
                    ),
                    dp.LegendItem(
                        icon=WBColor.GREEN_LIGHT.value,
                        label=_("Equal"),
                        value="EQUAL",
                    ),
                ],
            ),
        ]

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        hide_rating = self.request.GET.get("answer_type") == ReviewQuestion.ANSWERTYPE.TEXT
        fields = [
            dp.Field(key="category_question_name", label=_("Category"), width=Unit.PIXEL(200), hide=True),
            dp.Field(key="question_name", label=_("Question"), width=Unit.PIXEL(780)),
        ]
        if self.view.review.moderator == self.view.request.user.profile:
            fields.append(dp.Field(key="weight", label=_("Weight")))
        fields += [
            dp.Field(key="reviewee", label=_("Reviewee"), hide=hide_rating),
            dp.Field(key="reviewer", label=_("Reviewer"), hide=hide_rating),
            dp.Field(key="peers", label=_("Peers"), hide=True),
            dp.Field(key="comment_reviewee", label=_("Reviewee"), width=Unit.PIXEL(450)),
            dp.Field(key="comment_reviewer", label=_("Reviewer"), width=Unit.PIXEL(450)),
            dp.Field(key="comment_peers", label=_("Peers"), width=Unit.PIXEL(450), hide=True),
        ]
        return (
            dp.ListDisplay(fields=fields)
            if hide_rating
            else dp.ListDisplay(fields=fields, formatting=self._get_formatting_rules(), legends=self._get_legends())
        )


class ReviewProgressPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        _fields = [
            dp.Field(key="review_name", label=_("Review"), width=Unit.PIXEL(800)),
            dp.Field(key="reviewee", label=_("Reviewee")),
            dp.Field(key="reviewer", label=_("Reviewer")),
            dp.Field(key="peers", label=_("Peers")),
        ]
        return dp.ListDisplay(fields=_fields)



from datetime import date
from unittest.mock import patch

import pandas as pd
import pytest
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker

from wbhuman_resources.models import AbsenceRequest
from wbhuman_resources.models.employee import EmployeeHumanResource
from wbhuman_resources.tasks import assign_balance, send_mail_to_accounting

fake = Faker()


@pytest.mark.django_db
class TestHumanResourceTasks:
    @pytest.mark.parametrize("accounting_email_1,accounting_email_2", [(fake.email(), fake.email())])
    @patch.object(EmployeeHumanResource, "get_end_of_month_employee_balance_report_df")
    def test_send_mail_to_accounting_recipient(self, mock_fct, user_factory, accounting_email_1, accounting_email_2):
        global_preferences_registry.manager()["wbhuman_resources__accounting_company_emails"] = (
            f"{accounting_email_1};{accounting_email_2}"
        )
        mock_fct.return_value = pd.DataFrame(data=["a", "b"])

        admin_user = user_factory.create(is_superuser=True)
        content_type = ContentType.objects.get_for_model(AbsenceRequest)
        permission = Permission.objects.get(
            codename="administrate_absencerequest",
            content_type=content_type,
        )
        admin_user.user_permissions.add(permission)
        user_factory.create(
            is_superuser=True
        )  # we create another superuser to check if administirators don't contain it

        send_mail_to_accounting()
        assert len(mail.outbox) == 1
        last_mail = mail.outbox[-1]

        assert set(last_mail.to) == {accounting_email_2, accounting_email_1}
        assert set(last_mail.cc) == {admin_user.email}

    @pytest.mark.parametrize("year_str", [(fake.year())])
    def test_assign_yearly_balance(self, year_str, employee_human_resource_factory):
        year = int(year_str)
        test_date = date(year, 1, 1)
        active_employee = employee_human_resource_factory.create(
            enrollment_at=test_date,
            is_active=True,
            occupancy_rate=1,
            extra_days_frequency=EmployeeHumanResource.ExtraDaysBalanceFrequency.YEARLY,
        )
        unactive_employee = employee_human_resource_factory.create(
            enrollment_at=test_date,
            is_active=False,
            occupancy_rate=1,
            extra_days_frequency=EmployeeHumanResource.ExtraDaysBalanceFrequency.YEARLY,
        )
        assert not active_employee.balances.exists()
        assign_balance(test_date)
        assert not unactive_employee.balances.exists()
        assert active_employee.balances.count() == 1  # check that we get only this year balance
        employee_balance = active_employee.balances.get(year=test_date.year)
        allowance = employee_balance.monthly_allowances.first()
        assert (
            employee_balance.monthly_allowances.count() == 1
        )  # check that a unique allowance was granted to the employee
        assert (
            allowance.hourly_allowance == 25 * 8
        )  # check the the hourly amount correspond ot the base 25 days per year in hours



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class DayOffDayOffCalendarEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:calendar-dayoff-list", args=[self.view.kwargs["calendar_id"]], request=self.request
        )


class DefaultDailyPeriodDayOffCalendar(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:calendar-defaultperiod-list",
            args=[self.view.kwargs["calendar_id"]],
            request=self.request,
        )



from django.db.models.signals import ModelSignal

# this signal gathers all activity report needed to be inserted into the daily brief
add_employee_activity_to_daily_brief = ModelSignal(use_caching=True)



from datetime import datetime, timedelta
from decimal import Decimal
from typing import TypeVar

from celery import shared_task
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.core.files.base import ContentFile
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.utils import timezone
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from slugify import slugify
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import Person
from wbcore.contrib.documents.models import Document, DocumentType
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.markdown.utils import custom_url_fetcher
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from wbcore.models.fields import YearField
from wbcore.models.orderable import OrderableModel
from wbcore.utils.models import CloneMixin, ComplexToStringMixin
from weasyprint import HTML

from wbhuman_resources.models.employee import get_main_company
from wbhuman_resources.models.kpi import Evaluation

SelfReview = TypeVar("SelfReview", bound="Review")
User = get_user_model()


def can_trigger_review(instance, user):
    return user.profile == instance.moderator and not instance.is_template and instance.reviewee and instance.reviewer


def can_validate_review(instance, user):
    return can_trigger_review(instance, user) and instance.signed_reviewee and instance.signed_reviewer


class ReviewGroup(WBModel):
    class Meta:
        verbose_name = _("Review Group")
        verbose_name_plural = _("Review Groups")

    name = models.CharField(max_length=255, verbose_name=_("Name"))
    employees = models.ManyToManyField("directory.Person", related_name="reviewgroups", blank=True)

    def __str__(self):
        return f"{self.name}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewgroup"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewgrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class Review(CloneMixin, ComplexToStringMixin, WBModel):
    class Status(models.TextChoices):
        PREPARATION_OF_REVIEW = "PREPARATION_OF_REVIEW", _("Stage 1: Preparation of review")
        FILL_IN_REVIEW = "FILL_IN_REVIEW", _("Stage 2: Fill in review")
        REVIEW = "REVIEW", _("Stage 3: Review")
        EVALUATION = "EVALUATION", _("Stage 4: Evalutation")
        VALIDATION = "VALIDATION", _("Stage 5: Validation")

        @classmethod
        def get_color_map(cls):
            colors = [
                WBColor.BLUE_LIGHT.value,
                "rgb(230,230,250)",
                WBColor.YELLOW_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
                WBColor.GREEN_DARK.value,
            ]
            return [choice for choice in zip(cls, colors, strict=False)]

    class Type(models.TextChoices):
        ANNUAL = "ANNUAL", _("Annual")
        INTERMEDIARY = "INTERMEDIARY", _("Intermediary")

    class Meta:
        verbose_name = _("Review")
        verbose_name_plural = _("Reviews")
        permissions = [
            ("administrate_review", "Can Administrate Reviews"),
        ]

        notification_types = [
            create_notification_type(
                code="wbhuman_resources.review.notify",
                title="Review Notification",
                help_text="Notifies you when a review has been submitted",
            )
        ]

    from_date = models.DateField(null=True, blank=True, verbose_name=_("From"))
    to_date = models.DateField(null=True, blank=True, verbose_name=_("To"))
    review_deadline = models.DateField(null=True, blank=True, verbose_name=_("Deadline"))
    review = models.DateTimeField(null=True, blank=True, verbose_name=_("Review Date"))
    auto_apply_deadline = models.BooleanField(default=True, verbose_name=_("Auto Apply Deadline"))
    year = YearField(null=True, blank=True, verbose_name=_("Year"))
    type = models.CharField(max_length=32, choices=Type.choices, default=Type.ANNUAL, verbose_name=_("Type"))
    status = FSMField(
        default=Status.PREPARATION_OF_REVIEW,
        choices=Status.choices,
        verbose_name=_("Status"),
        help_text=_("Indicates one of the four stages defined by the workflow"),
    )

    reviewee = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="reviewee_reviews",
        on_delete=models.deletion.SET_NULL,
        verbose_name=_("Reviewee"),
    )
    reviewer = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="reviewer_reviews",
        on_delete=models.deletion.SET_NULL,
        verbose_name=_("Reviewer"),
    )

    moderator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="moderator_reviews",
        on_delete=models.deletion.SET_NULL,
    )

    review_group = models.ForeignKey(
        to="wbhuman_resources.ReviewGroup",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="review_related",
        verbose_name=_("Group"),
    )

    is_template = models.BooleanField(default=False)
    feedback_reviewee = models.TextField(default="", blank=True, verbose_name=_("Feedback Reviewee"))
    feedback_reviewer = models.TextField(default="", blank=True, verbose_name=_("Feedback Reviewer"))

    signed_reviewee = models.DateTimeField(null=True, blank=True, verbose_name=_("Date of reviewee's signature"))
    signed_reviewer = models.DateTimeField(null=True, blank=True, verbose_name=_("Date of reviewer's signature"))

    completely_filled_reviewee = models.DateTimeField(
        null=True, blank=True, verbose_name=_("Completely Filled Out Reviewee")
    )
    completely_filled_reviewer = models.DateTimeField(
        null=True, blank=True, verbose_name=_("Completely Filled Out Reviewer")
    )

    changed = models.DateTimeField(auto_now=True, null=True, blank=True)

    def compute_str(self) -> str:
        _str = ""
        if self.is_template:
            _str += (
                _("{group} [Template]").format(group=str(self.review_group)) if self.review_group else _("[Template]")
            )
        else:
            _str += (
                _("{reviewee}'s Review").format(reviewee=self.reviewee.computed_str) if self.reviewee else _("Review")
            )
        if self.from_date and self.to_date:
            _str += f" - ({self.from_date} - {self.to_date})"
        elif self.year:
            _str += f" - ({self.year})"
        return _str

    def __str__(self):
        return self.computed_str

    def save(self, *args, **kwargs):
        if self.reviewee and (employee := getattr(self.reviewee, "human_resources", None)) and not self.reviewer:
            self.reviewer = next(employee.get_managers(only_direct_manager=True), None)
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    @property
    def is_cloneable(self) -> bool:
        """
        Property used by the CloneMixin to disable the instance cloning functionality from the view
        """
        return self.is_template

    def _clone(self, **kwargs) -> SelfReview:
        """
        Create new row in database with the same data as original instance have.
        We need to clone also the related objects and set correct foreign key value
        """
        object_copy = Review.objects.get(id=self.id)
        object_copy.id = None
        object_copy.save()
        for related_object in self.questions.all():
            related_object.review = object_copy
            related_object.id = None
            related_object.save()
        return object_copy

    @classmethod
    def get_administrators(cls) -> QuerySet[User]:
        return (
            get_user_model()
            .objects.filter(
                Q(groups__permissions__codename="administrate_review")
                | Q(user_permissions__codename="administrate_review")
            )
            .distinct()
        )

    @transition(
        field=status,
        source=Status.PREPARATION_OF_REVIEW,
        target=Status.FILL_IN_REVIEW,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label=_("Submit"),
                action_label=_("Stage 2: Fill in review"),
                description_fields=_(
                    "<p>Status: <b>Stage 1: Preparation of review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to submit this review to <b>Stage 2: Fill in review?</b></p>"
                ),
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        submit_review.delay(self.id)

    @transition(
        field=status,
        source=Status.FILL_IN_REVIEW,
        target=Status.REVIEW,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="finalize",
                label=_("Finalize"),
                action_label=_("Stage 3: Review"),
                description_fields=_(
                    "<p>Status: <b>Stage 2: Fill in review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to send this review to <b>Stage 3: Review?</b></p>"
                ),
            )
        },
    )
    def finalize(self, by=None, description=None, **kwargs):
        if not self.review_deadline:
            self.review_deadline = datetime.now().date()
        finalize_review.delay(self.id)
        if not by:
            self.send_review_notification(
                title=gettext("Stage 3: Review - {self}").format(self=str(self)),
                message=gettext("{self} has moved to stage 3: Review. You can now organize the evaluation").format(
                    self=str(self)
                ),
                recipient=self.moderator.user_account,
                message_alert_deadline=False,
            )

    @transition(
        field=status,
        source=Status.REVIEW,
        target=Status.FILL_IN_REVIEW,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.EDIT.icon,
                key="undo",
                label=_("Undo"),
                action_label=_("Stage 2: Fill in review"),
                description_fields=_(
                    "<p>Status: <b>Stage 3: Review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to send this review to <b>Stage 2: Fill in review?</b></p>"
                ),
            )
        },
    )
    def undo(self, by=None, description=None, **kwargs):
        if self.review_deadline <= datetime.now().date():
            self.review_deadline = datetime.now().date() + timedelta(days=1)

    @transition(
        field=status,
        source=Status.REVIEW,
        target=Status.EVALUATION,
        permission=can_trigger_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="evaluate",
                label=_("Evaluate"),
                action_label=_("Stage 4: Evaluation"),
                description_fields=_(
                    "<p>Status: <b>Stage 3: Review</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Deadline: <b>{{review_deadline}}</b></p>  <p>Do you want to send this review to <b>Stage 4: Evaluation?</b></p>"
                ),
            )
        },
    )
    def evaluate(self, by=None, description=None, **kwargs):
        if not self.review:
            self.review = timezone.now()

    @transition(
        field=status,
        source=Status.EVALUATION,
        target=Status.VALIDATION,
        permission=can_validate_review,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:review",),
                icon=WBIcon.SEND.icon,
                key="validation",
                label=_("Validate"),
                action_label=_("Stage 4: Evaluation"),
                description_fields=_(
                    "<p>Status: <b>Stage 4: Evaluation</b></p> <p>Reviewee: <b>{{_reviewee.computed_str}}</b></p>  <p>Reviewer: <b>{{_reviewer.computed_str}}</b></p>  <p>From: <b>{{from_date}}</b></p> <p>To: <b>{{to_date}}</b></p> <p>Do you want to send this review to <b>Stage 5: Validation?</b></p>"
                ),
            )
        },
    )
    def validation(self, by=None, description=None, **kwargs):
        users = (
            get_user_model()
            .objects.filter(
                Q(groups__permissions__codename="administrate_review")
                | Q(user_permissions__codename="administrate_review")
            )
            .distinct()
        )
        for user in users:
            send_review_report_via_mail.delay(user.id, self.id)

    def send_review_notification(self, title, message, recipient, message_alert_deadline=True):
        if message_alert_deadline:
            message += gettext(
                "<p>Please pay particular attention to the deadline <b>{deadline}</b>, the review will automatically move to step 3 of the workflow on that date. At this stage the data will be frozen, you will not be able to modify it.</p>"
            ).format(deadline=self.review_deadline)

        send_notification(
            code="wbhuman_resources.review.notify",
            title=title,
            body=message,
            user=recipient,
            reverse_name="wbhuman_resources:review-detail",
            reverse_args=[self.id],
        )

    def generate_pdf(self):
        html = get_template("review/review_report.html")

        text_answers = ReviewAnswer.objects.filter(
            Q(question__review=self) & Q(question__answer_type=ReviewQuestion.ANSWERTYPE.TEXT)
        ).order_by("question__category__order", "question__order")
        category_ids = text_answers.values_list("question__category", flat=True).distinct()

        table = {}
        for category_id in category_ids:
            if category_id:
                category = ReviewQuestionCategory.objects.get(id=category_id)
                table[category.id] = {"name": category.name, "questions": {}}
            else:
                category_id = ""
                category = None
                table[""] = {"name": "", "questions": {}}
            question_ids = (
                text_answers.filter(question__category=category)
                .order_by("question__order")
                .values_list("question", flat=True)
                .distinct()
            )
            for question_id in question_ids:
                question = ReviewQuestion.objects.get(id=question_id)
                table[category_id]["questions"][question_id] = {"name": question.question, "answers": {}}
                for answer in text_answers.filter(
                    Q(question__category=category) & Q(question__id=question_id)
                ).order_by("question__order", "answered_by"):
                    table[category_id]["questions"][question_id]["answers"][answer.answered_by] = answer.answer_text

        rating_answers = (
            ReviewAnswer.objects.filter(Q(question__review=self) & Q(question__answer_type="RATING"))
            .annotate(
                is_reviewee=Case(
                    When(answered_by=self.reviewee, then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
                is_reviewer=Case(
                    When(answered_by=self.reviewer, then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
            )
            .order_by("question__category__order")
        )
        category_ids = rating_answers.values_list("question__category", flat=True).distinct()

        rating_table = {}
        for category_id in category_ids:
            if category_id:
                category = ReviewQuestionCategory.objects.get(id=category_id)
                rating_table[category.id] = {"name": category.name, "questions": {}}
            else:
                category_id = ""
                category = None
                rating_table[""] = {"name": "", "questions": {}}
            question_ids = (
                rating_answers.filter(question__category=category)
                .order_by("question__order")
                .values_list("question", flat=True)
                .distinct()
            )
            for question_id in question_ids:
                question = ReviewQuestion.objects.get(id=question_id)
                rating_table[category_id]["questions"][question_id] = {
                    "name": question.question,
                    "answers": {},
                    "answers_text": {},
                }
                for answer in rating_answers.filter(
                    Q(question__category=category) & Q(question__id=question_id)
                ).order_by("question__order", "answered_by"):
                    if answer.is_reviewee:
                        rating_table[category_id]["questions"][question_id]["answers"]["reviewee"] = (
                            answer.answer_number if answer.answer_number else "-"
                        )
                        rating_table[category_id]["questions"][question_id]["answers_text"]["reviewee"] = (
                            answer.answer_text if answer.answer_text else None
                        )
                    if answer.is_reviewer:
                        rating_table[category_id]["questions"][question_id]["answers"]["reviewer"] = (
                            answer.answer_number if answer.answer_number else "-"
                        )
                        rating_table[category_id]["questions"][question_id]["answers_text"]["reviewer"] = (
                            answer.answer_text if answer.answer_text else None
                        )

        total_reviewee = rating_answers.filter(is_reviewee=True).aggregate(total=Sum("question__weight"))["total"]
        total_reviewer = rating_answers.filter(is_reviewer=True).aggregate(total=Sum("question__weight"))["total"]
        global_rating_reviewee = (
            rating_answers.filter(is_reviewee=True)
            .annotate(global_rating=F("question__weight") * F("answer_number"))
            .aggregate(total=Sum("global_rating"))["total"]
        )
        global_rating_reviewer = (
            rating_answers.filter(is_reviewer=True)
            .annotate(global_rating=F("question__weight") * F("answer_number"))
            .aggregate(total=Sum("global_rating"))["total"]
        )
        global_rating_reviewee = (
            0 if total_reviewee is None or global_rating_reviewee is None else global_rating_reviewee / total_reviewee
        )
        global_rating_reviewer = (
            0 if total_reviewer is None or global_rating_reviewer is None else global_rating_reviewer / total_reviewer
        )

        html_content = html.render(
            {
                "base_url": Site.objects.get_current().domain,
                "review": self,
                "table": table,
                "rating_table": rating_table,
                "global_rating_reviewee": round(global_rating_reviewee, 2),
                "global_rating_reviewer": round(global_rating_reviewer, 2),
                "number_total_rating": ReviewQuestion.objects.filter(Q(review=self) & Q(answer_type="RATING")).count(),
            }
        )
        return HTML(
            string=html_content, base_url=settings.BASE_ENDPOINT_URL, url_fetcher=custom_url_fetcher
        ).write_pdf()

    @classmethod
    def get_subquery_review_related_to(cls, requester):
        answers = ReviewAnswer.objects.filter(question__review=OuterRef("pk"), answered_by=requester).order_by()
        return Coalesce(Subquery(answers.values("question__review")[:1]), None)

    @classmethod
    def subquery_global_rating(cls, requester):
        answers = ReviewAnswer.objects.filter(
            question__review=OuterRef("pk"),
            answered_by=requester,
            question__answer_type=ReviewQuestion.ANSWERTYPE.RATING,
            answer_number__isnull=False,
        ).order_by()

        subquery = (
            answers.annotate(
                rating_weighted=F("question__weight") * F("answer_number"),
            )
            .values("question__review")
            .annotate(
                total_rating=Sum("rating_weighted"),
                total_weight=Sum("question__weight"),
            )
            .annotate(global_rating=Coalesce(F("total_rating") / F("total_weight"), Decimal(0)))
            .values("global_rating")[:1]
        )

        return Coalesce(Subquery(subquery), None)

    def get_question_categories(self) -> models.QuerySet:
        return ReviewQuestionCategory.objects.filter(questions_related__review=self).order_by("order").distinct()

    def get_answer_categories_for_user(self, user: User) -> models.QuerySet:
        if self.status in [self.Status.FILL_IN_REVIEW, self.Status.REVIEW]:
            return (
                ReviewQuestionCategory.objects.filter(
                    questions_related__answers__answered_by=user.profile, questions_related__review=self
                )
                .order_by("order")
                .distinct()
            )
        else:  # EVALUATION
            return (
                ReviewQuestionCategory.objects.filter(
                    questions_related__answer_type=ReviewQuestion.ANSWERTYPE.TEXT, questions_related__review=self
                )
                .order_by("order")
                .distinct()
            )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:review"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


class ReviewQuestionCategory(OrderableModel, WBModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"), unique=True)
    weight = models.DecimalField(default=1.0, max_digits=16, decimal_places=1, verbose_name=_("Weight"))

    class Meta(OrderableModel.Meta):
        verbose_name = _("Review Question Category")
        verbose_name_plural = _("Review Question Categories")

    def __str__(self):
        return f"{self.name}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewquestioncategory"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewquestioncategoryrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class ReviewQuestion(ComplexToStringMixin, OrderableModel, WBModel):
    order_with_respect_to = ("review", "category")

    class ANSWERTYPE(models.TextChoices):
        TEXT = "TEXT", _("Text")
        RATING = "RATING", _("Rating")

    review = models.ForeignKey(
        to="wbhuman_resources.Review",
        on_delete=models.CASCADE,
        related_name="questions",
        verbose_name=_("Review"),
    )
    category = models.ForeignKey(
        to="wbhuman_resources.ReviewQuestionCategory",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="questions_related",
        verbose_name=_("Category"),
    )
    question = models.TextField(default="", blank=True, verbose_name=_("Question"))
    mandatory = models.BooleanField(default=True, verbose_name=_("Mandatory"))
    answer_type = models.CharField(
        max_length=32,
        default=ANSWERTYPE.TEXT,
        choices=ANSWERTYPE.choices,
        verbose_name=_("Type"),
    )
    for_reviewee = models.BooleanField(default=True, verbose_name=_("For Reviewee"))
    for_reviewer = models.BooleanField(default=True, verbose_name=_("For Reviewer"))
    for_department_peers = models.BooleanField(default=False, verbose_name=_("For Department Peers"))
    for_company_peers = models.BooleanField(default=False, verbose_name=_("For Company Peers"))
    weight = models.DecimalField(default=1.0, max_digits=16, decimal_places=1, verbose_name=_("Weight"))
    evaluation = models.ForeignKey(
        to="wbhuman_resources.Evaluation",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="evaluation_questions",
        verbose_name=_("Evaluation"),
    )

    class Meta:
        verbose_name = _("Review Question")
        verbose_name_plural = _("Review Questions")
        ordering = ("review", "category", "order")

    def compute_str(self) -> str:
        return f"{self.question}"

    def __str__(self):
        return f"({self.id}) - {self.computed_str}"

    def save(self, *args, **kwargs):
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewquestion"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbhuman_resources:reviewquestionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


class ReviewAnswer(models.Model):
    class Meta:
        verbose_name = _("Review Answer")
        verbose_name_plural = _("Review Answers")

    question = models.ForeignKey(
        to="wbhuman_resources.ReviewQuestion",
        on_delete=models.CASCADE,
        related_name="answers",
        verbose_name=_("Question"),
    )

    answered_by = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="related_answers",
        on_delete=models.deletion.SET_NULL,
        verbose_name=_("Answered By"),
    )

    answered_anonymized = models.CharField(
        null=True, blank=True, max_length=255, verbose_name=_("Answered Anonymized")
    )
    answer_number = models.PositiveIntegerField(
        null=True, blank=True, verbose_name=_("Rating"), validators=[MinValueValidator(1), MaxValueValidator(4)]
    )
    answer_text = models.TextField(null=True, blank=True, verbose_name=_("Comment"))

    def __str__(self) -> str:
        return f"{self.question} - {self.answered_by}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbhuman_resources:reviewanswer"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "id"


@receiver(pre_save, sender=Review)
def pre_save_review(sender, instance, **kwargs):
    if (
        not instance.is_template
        and instance.auto_apply_deadline
        and instance.review_deadline
        and instance.status == Review.Status.FILL_IN_REVIEW
    ):
        if instance.review_deadline < datetime.now().date():
            instance.finalize()

    if (
        not instance.is_template
        and instance.signed_reviewee
        and instance.signed_reviewer
        and instance.status == Review.Status.EVALUATION
    ):
        instance.validation()


@shared_task
def finalize_review(review_id):
    review = Review.objects.get(id=review_id)
    ReviewAnswer.objects.filter(
        Q(question__review=review)
        & Q(answered_by=review.reviewee)
        & Q(question__answer_type=ReviewQuestion.ANSWERTYPE.RATING)
        & Q(answer_number=None)
    ).update(answer_number=1)


@shared_task
def submit_reviews_from_group(group_id, user_id):
    user = get_user_model().objects.get(id=user_id)

    reviews = Review.objects.filter(
        Q(review_group__id=group_id)
        & Q(status=Review.Status.PREPARATION_OF_REVIEW)
        & Q(moderator=user.profile)
        & Q(is_template=False)
    )
    for review in reviews:
        review.submit()
        review.save()


@shared_task
def create_review_from_template(
    template_id, from_date, to_date, review_deadline, auto_apply_deadline, employees, include_kpi
):
    template = Review.objects.get(id=template_id)

    from_date = datetime.strptime(from_date, "%Y-%m-%d").date() if from_date and from_date != "null" else None
    to_date = datetime.strptime(to_date, "%Y-%m-%d").date() if to_date and to_date != "null" else None
    review_deadline = (
        datetime.strptime(review_deadline, "%Y-%m-%d").date()
        if review_deadline and review_deadline != "null"
        else None
    )

    include_kpi = str(include_kpi).lower() in ("yes", "true", "t", "1") if include_kpi else False

    _auto_apply_deadline = (
        str(auto_apply_deadline).lower() in ("yes", "true", "t", "1")
        if auto_apply_deadline
        else template.auto_apply_deadline
    )
    list_employees = Person.objects.filter(id__in=employees.split(",")) if employees else []

    for employee in list_employees:
        review = Review.objects.create(
            from_date=from_date,
            to_date=to_date,
            review_deadline=review_deadline,
            reviewee=employee,
            moderator=template.moderator,
            auto_apply_deadline=_auto_apply_deadline,
            status=template.status,
            review_group=template.review_group,
            year=template.year,
            type=template.type,
        )

        for _question in ReviewQuestion.objects.filter(review__id=template.id).order_by("order"):
            kwargs = {
                "review": review,
                "category": _question.category,
                "question": _question.question,
                "answer_type": _question.answer_type,
                "mandatory": _question.mandatory,
                "for_reviewee": _question.for_reviewee,
                "for_reviewer": _question.for_reviewer,
                "for_department_peers": _question.for_department_peers,
                "for_company_peers": _question.for_company_peers,
                "order": _question.order,
                "weight": _question.weight,
            }
            ReviewQuestion.objects.create(**kwargs)

        if include_kpi:
            qs_evaluations = Evaluation.objects.filter(is_active=True)
            evaluations = (
                qs_evaluations.filter(
                    Q(person=employee) | (Q(person__isnull=True) & Q(kpi__evaluated_persons=employee))
                )
                .annotate(
                    goal=F("kpi__goal"),
                    kpi_name=F("kpi__name"),
                    interval=F("kpi__evaluated_intervals"),
                    period=F("kpi__period"),
                    person_name=F("person__computed_str"),
                )
                .order_by("kpi", "-evaluation_date")
                .distinct("kpi")
            )

            category, _ = ReviewQuestionCategory.objects.get_or_create(name="KPI Evaluations")
            for evaluation in evaluations:
                percentage = round((evaluation.evaluated_score / evaluation.goal) * 100, 2)
                person_name = evaluation.person_name if evaluation.person_name else "Group"
                _question = gettext(
                    """The KPI '{kpi_name}' is evaluated {interval} over the period from {lower_period} to {upper_period}.
{person_name}'s evaluation as of {evaluation_date} was {evaluated_score} for a goal of {goal}. That is an accomplished percentage of {percentage}%.
What do you think of this result?
                """
                ).format(
                    kpi_name=evaluation.kpi_name,
                    interval=evaluation.interval,
                    lower_period=evaluation.period.lower,
                    upper_period=evaluation.period.upper,
                    person_name=person_name,
                    evaluation_date=evaluation.evaluation_date,
                    evaluated_score=evaluation.evaluated_score,
                    goal=evaluation.goal,
                    percentage=percentage,
                )

                kwargs = {
                    "review": review,
                    "category": category,
                    "question": _question,
                    "answer_type": ReviewQuestion.ANSWERTYPE.RATING,
                    "evaluation": evaluation,
                }
                ReviewQuestion.objects.create(**kwargs)


@shared_task
def submit_review(review_id):
    review = Review.objects.get(id=review_id)
    questions = ReviewQuestion.objects.filter(review=review)
    dict_questions = {}

    questions_for_company = questions.filter(for_company_peers=True)
    if questions_for_company.exists() and (main_company_id := get_main_company()):
        if _company := review.reviewee.employers.filter(id=main_company_id).first():
            for _person in (
                Person.objects.filter_only_internal()
                .filter(employers=_company)
                .exclude(Q(id=review.reviewee.id) | Q(id=review.reviewer.id))
            ):
                dict_questions[_person] = questions_for_company

    questions_for_department = questions.filter(for_department_peers=True)
    if questions_for_department.exists() and hasattr(review.reviewee, "human_resources"):
        if review.reviewee.human_resources.position:
            employees_peers = []
            for _position in review.reviewee.human_resources.position.get_ancestors(include_self=True):
                employees_peers += _position.get_employees().exclude(
                    Q(id=review.reviewee.id) & Q(id=review.reviewer.id)
                )
            person_peers = [_employee.profile for _employee in list(set(employees_peers))]
            for _peer in list(set(person_peers)):
                if _peer in dict_questions:
                    dict_questions[_peer] = questions_for_department.union(questions_for_company)
                else:
                    dict_questions[_peer] = questions_for_department

    dict_questions[review.reviewee] = questions.filter(for_reviewee=True)
    dict_questions[review.reviewer] = questions.filter(for_reviewer=True)

    for person, values in dict_questions.items():
        for _question in values:
            kwargs = {"question": _question, "answered_by": person, "answered_anonymized": hash(person)}
            if _evaluation := _question.evaluation:
                rating = _evaluation.get_rating()
                kwargs.update({"answer_number": rating})
            ReviewAnswer.objects.create(**kwargs)
        if hasattr(person, "user_account"):
            if person == review.reviewee:
                msg = gettext(
                    "Dear {first_name} {last_name}, <p>your review is ready. You can fill it in until {deadline}.</p>"
                ).format(first_name=person.first_name, last_name=person.last_name, deadline=review.review_deadline)
            elif person == review.reviewer:
                msg = gettext(
                    "Dear {first_name} {last_name}, <p>you are the reviewer of the review: {review}. You can fill in the questions related to you until <b>{deadline}.</b></p>"
                ).format(
                    first_name=person.first_name,
                    last_name=person.last_name,
                    review=str(review),
                    deadline=review.review_deadline,
                )
            else:
                msg = gettext(
                    "Dear {first_name} {last_name}, <p>you are a peer of the review: {review}. You can fill in the questions related to you until <b>{deadline}.</b></p>"
                ).format(
                    first_name=person.first_name,
                    last_name=person.last_name,
                    review=str(review),
                    deadline=review.review_deadline,
                )

            review.send_review_notification(
                title=gettext("Stage 2: Fill in review - {review}").format(review=str(review)),
                message=msg,
                recipient=person.user_account,
            )


@shared_task
def send_review_report_via_mail(user_id, review_id):
    user = get_user_model().objects.get(id=user_id)
    review = Review.objects.get(id=review_id)
    pdf_content = review.generate_pdf()
    filename = f"{slugify(str(review))}.pdf"

    content_file = ContentFile(pdf_content, name=filename)
    document_type, _ = DocumentType.objects.get_or_create(name="mailing")

    document, _ = Document.objects.update_or_create(
        document_type=document_type,
        system_created=True,
        system_key=f"review-{review.id}-{filename}",
        defaults={
            "file": content_file,
            "name": filename,
            "permission_type": Document.PermissionType.PRIVATE,
            "creator": user,
        },
    )
    document.link(review)

    document.send_email(to_emails=user.email, as_link=True, subject=gettext("Review PDF - ") + review.computed_str)



from datetime import date, timedelta

import pandas as pd
import pytest
from faker import Faker

from wbhuman_resources.models import EmployeeHumanResource

fake = Faker()


@pytest.mark.django_db
class TestUtil:
    @pytest.mark.parametrize("test_date", [(fake.date_object())])
    def test_get_employee_absence_periods_df(
        self,
        employee_human_resource,
        absence_request_periods_factory,
        day_off_factory,
        employee_weekly_off_periods_factory,
        test_date,
    ):
        # Basic utility coverage test. The underlying functions are already tested.
        p = absence_request_periods_factory.create(date=test_date, employee=employee_human_resource)
        w = employee_weekly_off_periods_factory.create(
            employee=employee_human_resource, weekday=(test_date + timedelta(days=1)).weekday()
        )
        day_off_factory.create(calendar=employee_human_resource.calendar, date=test_date + timedelta(days=2))
        res = EmployeeHumanResource.get_employee_absence_periods_df(
            employee_human_resource.calendar,
            (test_date - pd.tseries.offsets.Week(1)).date(),
            (test_date + pd.tseries.offsets.Week(1)).date(),
            EmployeeHumanResource.active_employees.all(),
        ).set_index(["employee", "period", "date", "type"])
        assert not res.loc[(employee_human_resource.id, p.default_period.id, test_date, "Vacation"), :].empty
        assert not res.loc[
            (employee_human_resource.id, w.period.id, test_date + timedelta(days=1), "Day Off"), :
        ].empty
        assert not res.loc[
            (
                employee_human_resource.id,
                employee_human_resource.calendar.default_periods.earliest("lower_time").id,
                test_date + timedelta(days=2),
                "Holiday",
            ),
            :,
        ].empty
        assert not res.loc[
            (
                employee_human_resource.id,
                employee_human_resource.calendar.default_periods.latest("lower_time").id,
                test_date + timedelta(days=2),
                "Holiday",
            ),
            :,
        ].empty
        with pytest.raises(KeyError):
            assert not res.loc[(employee_human_resource.id, 2, test_date + timedelta(days=3), "Holiday"), :].empty

    @pytest.mark.parametrize(
        "end_of_month_date",
        [
            (fake.date_between(date(2021, 1, 1), date(2021, 3, 31))),
            (fake.date_between(date(2021, 4, 1), date(2021, 12, 31))),
        ],
    )
    def test_get_end_of_month_employee_balance_report_df(
        self,
        employee_human_resource,
        employee_year_balance_factory,
        balance_hourly_allowance_factory,
        absence_request_periods_factory,
        end_of_month_date,
    ):
        end_of_month_date = (end_of_month_date + pd.tseries.offsets.MonthEnd(0)).date()
        previous_balance = employee_year_balance_factory.create(
            employee=employee_human_resource, year=end_of_month_date.year - 1, extra_balance=0
        )
        balance_hourly_allowance_factory.create(balance=previous_balance)

        current_balance = employee_year_balance_factory.create(
            employee=employee_human_resource, year=end_of_month_date.year, extra_balance=0
        )
        balance_hourly_allowance_factory.create(balance=current_balance)
        previous_period = absence_request_periods_factory.create(
            balance=previous_balance,
            employee=employee_human_resource,
            date=fake.date_between(date(end_of_month_date.year, 1, 1), end_of_month_date - timedelta(days=1)),
        )
        current_period = absence_request_periods_factory.create(
            balance=current_balance,
            employee=employee_human_resource,
            date=fake.date_between(previous_period.date, end_of_month_date),
        )

        res = EmployeeHumanResource.get_end_of_month_employee_balance_report_df(
            EmployeeHumanResource.active_employees.all(), end_of_month_date
        ).to_dict("records")[0]
        expected_total_balance = (
            current_balance.balance_in_days
            + previous_balance.total_vacation_hourly_balance_in_days
            + previous_balance.total_vacation_hourly_usage_in_days
        )
        expected_usage = (previous_period.total_hours + current_period.total_hours) / 8
        assert res == {
            "Employee": employee_human_resource.computed_str,
            "Total Balance": expected_total_balance,
            "End of Month Usage": expected_usage,
            "Available Balance End of Month": expected_total_balance - expected_usage,
        }



from django.contrib import admin

from ..models.employee import (
    BalanceHourlyAllowance,
    EmployeeHumanResource,
    EmployeeYearBalance,
    Position,
)
from .calendars import EmployeeWeeklyOffPeriodsInLine


class PositionInline(admin.TabularInline):
    model = Position
    extra = 0


@admin.register(Position)
class PositionAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("name", "height", "manager")

    raw_id_fields = ["groups", "manager"]
    autocomplete_fields = ["groups", "manager"]


class BalanceHourlyAllowanceTabularAdmin(admin.TabularInline):
    model = BalanceHourlyAllowance
    fk_name = "balance"


@admin.register(EmployeeYearBalance)
class EmployeeYearBalanceAdmin(admin.ModelAdmin):
    inlines = [BalanceHourlyAllowanceTabularAdmin]


class EmployeeYearBalanceInline(admin.TabularInline):
    model = EmployeeYearBalance
    fields = (
        "year",
        "balance",
        "daily_hours",
        "number_mandatory_days_off",
        "total_vacation_hourly_usage",
        "total_vacation_hourly_balance",
        "balance_in_days",
        "number_mandatory_days_off_in_days",
        "total_vacation_hourly_usage_in_days",
        "total_vacation_hourly_balance_in_days",
    )
    readonly_fields = [
        "balance",
        "daily_hours",
        "number_mandatory_days_off",
        "total_vacation_hourly_usage",
        "total_vacation_hourly_balance",
        "balance_in_days",
        "number_mandatory_days_off_in_days",
        "total_vacation_hourly_usage_in_days",
        "total_vacation_hourly_balance_in_days",
    ]
    extra = 0
    ordering = ("-year",)
    show_change_link = True

    def _number_mandatory_days_off(self, obj):
        return obj._number_mandatory_days_off

    def _total_vacation_hourly_usage(self, obj):
        return obj._total_vacation_hourly_usage

    def _total_vacation_hourly_balance(self, obj):
        return obj._total_vacation_hourly_balance

    def _balance_in_days(self, obj):
        return obj._balance_in_days

    def _number_mandatory_days_off_in_days(self, obj):
        return obj._number_mandatory_days_off_in_days

    def _total_vacation_hourly_usage_in_days(self, obj):
        return obj._total_vacation_hourly_usage_in_days

    def _total_vacation_hourly_balance_in_days(self, obj):
        return obj._total_vacation_hourly_balance_in_days


@admin.register(EmployeeHumanResource)
class EmployeeHumanResourceAdmin(admin.ModelAdmin):
    fieldsets = (
        (
            "",
            {
                "fields": (
                    "profile",
                    "is_active",
                    "extra_days_frequency",
                    "occupancy_rate",
                    "contract_type",
                    "position",
                    "enrollment_at",
                    "calendar",
                )
            },
        ),
    )
    raw_id_fields = ("profile",)
    search_fields = ("profile__computed_str",)
    list_display = ("profile", "contract_type", "position", "calendar")
    inlines = (EmployeeYearBalanceInline, EmployeeWeeklyOffPeriodsInLine)
